from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import sys, os

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import clip_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.training import optimizer as tf_optimizer
from tensorflow.python.training import training_util

class deconvModel(object):
  def __init__(self, FLAGS, VAE, datasets, input_size, output_size, num_samples, reconstruction=None, loadings=None, marker_gene_masks=None, hvg_masks=None, scope="prop_model"):
    ## Params for deconvolution model
    self.input_size  = input_size
    self.output_size = output_size
    self.num_samples = num_samples
    self.batch_size  = np.amin((FLAGS.batch_size_mixture, self.num_samples)) ## Batch size should not be larger than dataset size
    ## Marker gene mask
    self.marker_gene_mask = None
    ## Organizing name for cmobined model
    self.scope = scope
    ## Cell types in model
    self.current_celltypes = np.unique(datasets)
    self.num_component = len(self.current_celltypes)
    ##
    self.step = tf.Variable(1, name='prop_step', trainable=False, dtype=tf.int32)

    ## Build combined model
    with tf.variable_scope(self.scope, tf.AUTO_REUSE):
        self.build_dataset();
        self.build_model(FLAGS);
        self.deconvolution(FLAGS, VAE);
        ## MSE between reconstruction and measured mixture
        self.add_prop_loss(self.data_batch, self.mixture_rec)
        ## Create training operation
        self.create_train_op(FLAGS)
        self.create_diagnostics(FLAGS)

  def build_dataset(self):
    with tf.name_scope("mixture_data"):
        ## Mixture indexing
        self.index = tf.range(start=0, limit=self.num_samples, dtype=tf.int32)
        ## Dataset
        self.mix_data_ph = tf.placeholder(tf.float32, (None, self.input_size))
        self.mix_dataset = tf.data.Dataset.from_tensor_slices((self.mix_data_ph, self.index)).shuffle(self.num_samples).repeat().batch(self.batch_size)
        self.mix_iter_data  = self.mix_dataset.make_initializable_iterator()
        self.mix_data_batch, self.data_index  = self.mix_iter_data.get_next()

  def build_model(self, FLAGS):
    with tf.name_scope("mixture_proportions"):
        ## Mixture weights
        self.mixture_weights = tf.get_variable(name="prop_coef",
                                               shape=[self.num_samples, self.num_component],
                                               dtype=tf.float32,
                                               initializer=tf.zeros_initializer,
                                               regularizer=tf.contrib.layers.l1_regularizer(FLAGS.l1_reg_weight) if (FLAGS.mix_weight_reg is True) else None,
                                               trainable=True)
        ## Normalize mixture weights
        self.proportions = tf.nn.softmax(self.mixture_weights, axis=-1, name="proportions")
    ## Hold results of linear reconstitution of mixture
    self.mixture_rec = tf.zeros([self.batch_size, self.output_size], tf.float32)

  def deconvolution(self, FLAGS, VAE):
    ## Compute mixture from weighted components
    for index, value in enumerate(self.current_celltypes, 0):
        #with tf.name_scope(value):
        with tf.variable_scope(value, tf.AUTO_REUSE):
            data_emb = VAE[value].encoder_func(self.mix_data_batch, is_training=False).sample(FLAGS.num_monte_carlo)
            data_rec = tf.reduce_mean(VAE[value].decoder_func(data_emb, is_training=False).mean(), axis=0)
            ## TPM specific reconstruction
            if FLAGS.tpm_softmax is True:
                data_rec = tf.nn.softmax(data_rec, axis=-1) * FLAGS.tpm_scale_factor
            ## Component specific contribution to mixture profile
            pure_data = tf.multiply(data_rec, tf.gather(self.proportions[:,index,tf.newaxis], self.data_index, axis=0))
        ## Proportion specific reconstructions, could be masked
        if VAE[value].marker_mask is not None:
            self.mixture_rec = tf.add(self.mixture_rec, tf.multiply(pure_data, VAE[value].marker_mask))
            self.data_batch = tf.multiply(self.mix_data_batch, VAE[value].marker_mask)
        else:
            self.mixture_rec = tf.add(self.mixture_rec, pure_data)
            self.data_batch = self.mix_data_batch
        tf.summary.histogram(value + "_weights", self.mixture_weights[:,index])
        tf.summary.histogram(value + "_probs", self.proportions[:,index])

  def add_prop_loss(self, data, rec):
    ## Define decoder reconstruction loss
    self.mse_mixture = tf.losses.mean_squared_error(data,
                                                    rec,
                                                    weights=1.0,
                                                    scope=self.scope)
    tf.summary.scalar("MSE", self.mse_mixture)

  def create_train_op(self, FLAGS):
    """Create and return training operation."""
    with tf.name_scope('Optimizer_prop'):
        ## Set up learning rate
        if FLAGS.decay_lr is True:
            with tf.name_scope("learning_rate_prop"):
                self.prop_learning_rate = tf.maximum(
                    tf.train.exponential_decay(
                        FLAGS.proportion_learning_rate,
                        self.step, ## Account for previous training
                        FLAGS.decay_step*10,
                        FLAGS.decay_rate),
                    FLAGS.min_learning_rate)
        else:
            self.prop_learning_rate = FLAGS.proportion_learning_rate

        ## Collect prop component loss
        self.proportion_loss = self.mse_mixture
        ## Minimize loss function
        optimizer = tf.train.AdamOptimizer(self.prop_learning_rate)
        self.train_proportion_op = optimizer.minimize(loss=self.proportion_loss, global_step=self.step, var_list=tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope))
        ## Monitor
        tf.summary.scalar('Learning_Rate', self.prop_learning_rate)
        tf.summary.scalar('Loss_Total', self.proportion_loss)
        return(self.train_proportion_op)

  def create_diagnostics(self, FLAGS):
    ## Diagnostic
    if FLAGS.combined_corr_check is True:
      self.corr = tfp.stats.correlation(self.mix_data_batch, self.mixture_rec, sample_axis=-1, event_axis=0)
      self.corr = tf.reduce_mean(tf.linalg.tensor_diag_part(self.corr), axis=0)
      tf.compat.v1.summary.scalar("Mixture_correlation", self.corr)



#! /usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys, os, glob, gzip, time
from functools import partial
from importlib import import_module

## Math
import numpy as np
import scipy as sp

## tensorflow imports
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.platform import app
from tensorflow.python.platform import flags

## Record evaluation metrics
class modelMetrics(object):
    def __init__(self, mode):
        self.step = {'train': [], 'test': [], 'train-post': []}
        self.loss = {'train': [], 'test': [], 'train-post': []}
        self.mse  = {'train': [], 'test': [], 'train-post': []}
        if mode is "VAE":
            self.log_probability = {'train': [], 'test': [], 'train-post': []}
            self.kl_divergence   = {'train': [], 'test': [], 'train-post': []}

## Record results and evaluation metrics
class deconvResult(object):
    def __init__(self, celltypeNames):
        self.proportions       = {}
        self.weights           = {}
        ## Reconstructions and embeddings
        self.deconv_data = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_emb  = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_logp = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_var  = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        #self.deconv_inv_data = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        ## Reconstructions and embeddings after correction
        self.deconv_data_post = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_emb_post  = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_logp_post = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        self.deconv_var_post  = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        #self.deconv_inv_data_post = {'component': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}, 'purified': {'train': dict.fromkeys(celltypeNames), 'test': dict.fromkeys(celltypeNames)}}
        ## Sampling
        self.deconv_samples_emb = {'prebatch': dict.fromkeys(celltypeNames), 'postbatch': dict.fromkeys(celltypeNames)}
        self.deconv_samples_rec = {'prebatch': dict.fromkeys(celltypeNames), 'postbatch': dict.fromkeys(celltypeNames)}
        ## Mixture reconstruction
        self.deconv_mixtures = {'prebatch': None, 'postbatch': None}
        ## Train/test metrics
        self.component_metrics  = dict([(key, modelMetrics(mode = "VAE")) for key in celltypeNames])
        self.proportion_metrics = modelMetrics(mode = "Mixture")
        self.mixture_metrics   = modelMetrics(mode = "Mixture")
        ## Model details
        self.celltypes        = celltypeNames
        self.results_dir      = None
        self.flags            = None
        ## mphate specifics
        self.m_phate_logging   = dict([(key, []) for key in celltypeNames])

    def update_component_metrics(self, mode, scope, step, loss, mse, logp, kl):
        self.component_metrics[scope].step[mode].append(step)
        self.component_metrics[scope].loss[mode].append(loss)
        self.component_metrics[scope].mse[mode].append(mse)
        self.component_metrics[scope].log_probability[mode].append(logp)
        self.component_metrics[scope].kl_divergence[mode].append(kl)

    def update_mixture_metrics(self, mode, loss, mse):
        self.proportion_metrics.loss[mode].append(loss)
        self.proportion_metrics.mse[mode].append(mse)

    def update_mixture_metrics(self, mode, loss, mse):
        self.mixture_metrics.loss[mode].append(loss)
        self.mixture_metrics.mse[mode].append(mse)

    def update_proportions(self, step, props):
        self.proportions[str(step)] = props
        #self.weights[str(step)]     = weights

    def update_mphate(self, scope, embedding):
        self.m_phate_logging[scope].append(embedding)

    def update_saver(self, datatype, mode, stage, scope, rec_data, embedding, log_prob, rec_var):
        if stage == "component":
            self.deconv_data[datatype][mode][scope] = rec_data
            self.deconv_emb[datatype][mode][scope]  = embedding
            self.deconv_logp[datatype][mode][scope] = log_prob
            self.deconv_var[datatype][mode][scope]  = rec_var
        elif stage == "combined":
            self.deconv_data_post[datatype][mode][scope] = rec_data
            self.deconv_emb_post[datatype][mode][scope]  = embedding
            self.deconv_logp_post[datatype][mode][scope] = log_prob
            self.deconv_var_post[datatype][mode][scope]  = rec_var
        else:
            print("Unknown stage")

    def sample_saver(self, datatype, scope, emb_samples, rec_samples):
        self.deconv_samples_emb[datatype][scope] = emb_samples
        self.deconv_samples_rec[datatype][scope] = rec_samples

    def mixture_saver(self, datatype, mixture_rec):
        self.deconv_mixtures[datatype] = mixture_rec

    def __repr__(self):
        return "Results of <NAME> deconvolution"

def initialize_model_dir(FLAGS):
    print("Running training method: " + FLAGS.training_method)
    ## Ensure save path exists
    FLAGS.logdir = os.path.abspath(FLAGS.logdir)
    ## Setup save path
    FLAGS.logdir = FLAGS.logdir + "/model_" + FLAGS.model
    print(FLAGS.logdir)
    ## Make path
    if not os.path.exists(os.path.abspath(FLAGS.logdir)):
       os.makedirs(os.path.abspath(FLAGS.logdir))
    return FLAGS

## Saves flags to a file
def write_flags(FLAGS):
    flag_dict = tf.app.flags.FLAGS.flag_values_dict()
    with open(FLAGS.logdir + '/run_flags.txt', 'w+') as file_out:
        [file_out.write('{0}\t{1}\n'.format(key, value)) for key, value in flag_dict.items()]

def summary_and_checkpoint(FLAGS, graph):
    ## Tensorboard and model checkpointing
    summary_op = tf.summary.merge_all()
    ## Write summaries
    summary_writer = tf.summary.FileWriter(os.path.abspath(FLAGS.logdir) +  "/train", graph, filename_suffix='-train')
    summary_writer_test = tf.summary.FileWriter(os.path.abspath(FLAGS.logdir) +  "/test", filename_suffix='-test')
    ## Model saving every X steps or Y hours
    saver = tf.train.Saver(max_to_keep=5,  keep_checkpoint_every_n_hours=1.0)
    return summary_op, summary_writer, summary_writer_test, saver

## Setup masking and annotations
def setup_annotations_masks(component_label, marker_gene_mask, hvg_mask, max_steps_component, celltypes):
    ## Convert old format to new dictionary for celltype labels from reference
    if type(component_label) is not dict:
        print("Converting component label list to dictionary")
        component_labels = {'standard_deconv' : component_label}
    else:
        component_labels = component_label

    ## Convert old format to dictionary for marker gene mask which is shared across cell types
    if type(marker_gene_mask) is not dict and marker_gene_mask is not None:
        print("Converting marker mask to dictionary")
        marker_gene_masks = {}
        for scope in celltypes:
            marker_gene_masks.update( {scope : marker_gene_mask} )
    else:
        marker_gene_masks = marker_gene_mask

    ## Convert old format to dictionary for marker gene mask which is shared across cell types
    if type(hvg_mask) is not dict and hvg_mask is not None:
        print("Converting hvg mask to dictionary")
        hvg_masks = {}
        for scope in celltypes:
            hvg_masks.update( {scope : hvg_mask} )
    else:
        hvg_masks = hvg_mask

    ## Convert old format to dictionary for marker gene mask which is shared across cell types
    if type(max_steps_component) is not dict and max_steps_component is not None:
        print("Converting hvg mask to dictionary")
        max_steps_components = {}
        for scope in celltypes:
            max_steps_components.update( {scope : max_steps_component} )
    else:
        max_steps_components = max_steps_component

    return component_labels['standard_deconv'], marker_gene_masks, hvg_masks, max_steps_components

def early_stopping(FLAGS, VAE, scope, corr):
    ## Early stopping if enabled and minimum number of steps has been achieved
    if corr > 0.85:
        VAE[scope].patience = VAE[scope].patience + 1
    elif corr > 0.90:
        VAE[scope].patience = VAE[scope].patience + 10
    else:
        VAE[scope].patience = VAE[scope].patience - 1

    if VAE[scope].patience >= FLAGS.max_patience:
        print("Early stopping triggered for: ", str(scope) + "   Corr: " + str(corr))
        VAE[scope].early_stop = True

def define_hardware(FLAGS):
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    config.log_device_placement = True
    config.allow_soft_placement = True
    config.graph_options.optimizer_options.global_jit_level = tf.OptimizerOptions.ON_1
    os.environ["CUDA_VISIBLE_DEVICES"] = FLAGS.cuda_device
    return config



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import sys, os

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import clip_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.platform import tf_logging as logging
from tensorflow.python.summary import summary
from tensorflow.python.training import monitored_session
from tensorflow.python.training import optimizer as tf_optimizer
from tensorflow.python.training import training_util

## Simple prior N(0,I)
def vae_prior(ndim,
              scope):
  with tf.name_scope('prior_Z'):
    return tfp.distributions.MultivariateNormalDiag(loc=tf.zeros(ndim),
                                                    scale_diag=tf.ones(ndim),
                                                    allow_nan_stats=False,
                                                    validate_args=True,
                                                    name="normal_prior_Z")

## Approximation: q(z|x)
def vae_encoder(inputs,
                ndim,
                FLAGS,
                scope,
                num_layers=3, ## This number excludes the embedding layer
                hidden_unit_2power=9, ## Specify the maximal number (2^X) hidden units.
                l2_weight=1e-4,
                dropout_rate=0.3,
                batch_norm=True,
                batch_renorm=False,
                is_training=True):
  with tf.name_scope('vae_encoder'):
      inputs = tf.cast(inputs, tf.float32)
      encoder = tf.layers.dense(inputs=inputs,
                                units=pow(2,hidden_unit_2power),
                                activation=tf.nn.relu,
                                kernel_initializer=tf.glorot_uniform_initializer(),
                                kernel_regularizer=tf.keras.regularizers.l2(l2_weight),
                                use_bias=True,
                                bias_initializer=init_ops.zeros_initializer(),
                                name='fc1')
      if batch_norm: encoder = tf.layers.batch_normalization(encoder, training=is_training, renorm=batch_renorm, name='batch_norm_1')
      encoder = tf.layers.dropout(inputs=encoder, rate=dropout_rate, training=is_training, name='drop1')
      ## Add additional layers while accounting for input layer
      for layer in range(1,num_layers):
          encoder = tf.layers.dense(inputs=encoder,
                                    units=pow(2,hidden_unit_2power-layer),
                                    activation=tf.nn.relu,
                                    kernel_initializer=tf.glorot_uniform_initializer(),
                                    kernel_regularizer=tf.keras.regularizers.l2(l2_weight),
                                    use_bias=True,
                                    bias_initializer=init_ops.zeros_initializer(),
                                    name='fc'+str(layer+1))
          if batch_norm: encoder = tf.layers.batch_normalization(encoder, training=is_training, renorm=batch_renorm, name='batch_norm_'+str(layer+1))
          encoder = tf.layers.dropout(inputs=encoder, rate=dropout_rate, training=is_training, name='drop'+str(layer+1))
      ## Embedding layer has user defined number of hidden units (ndim)
      encoder = tf.layers.dense(inputs=encoder,
                                units=2*ndim,
                                activation=None,
                                kernel_initializer=tf.glorot_uniform_initializer(),
                                kernel_regularizer=None,
                                use_bias=True,
                                bias_initializer=init_ops.zeros_initializer(),
                                name='fc'+str(num_layers+1))
      encoder = tfp.distributions.MultivariateNormalDiag(
            loc=encoder[..., :ndim],
            scale_diag=tf.nn.softplus(encoder[..., ndim:])+1e-8, ## Changed tf.nn.softplas due to numeric issues with large variances (Potentially unused dimensions)
            allow_nan_stats=False,
            validate_args=False,
            name="latent_Z")
  return encoder

## Reconstruction: p(x|z)
def vae_decoder(inputs,
                output_size,
                FLAGS,
                scope,
                l2_weight=1e-4,
                hidden_unit_2power=9,
                num_layers=3,
                dropout_rate=0.3,
                batch_norm=True,
                batch_renorm=False,
                is_training=True):
  with tf.name_scope('vae_decoder'):
      inputs = tf.cast(inputs, tf.float32)
      decoder = tf.layers.dense(inputs=inputs,
                                units=pow(2,(hidden_unit_2power-num_layers)+1),
                                activation=tf.nn.relu,
                                kernel_initializer=tf.glorot_uniform_initializer(),
                                kernel_regularizer=tf.keras.regularizers.l2(l2_weight),
                                use_bias=True,
                                bias_initializer=init_ops.zeros_initializer(),
                                name='fc1')
      if batch_norm: decoder = tf.layers.batch_normalization(decoder, training=is_training, renorm=batch_renorm, name='batch_norm_1')
      decoder = tf.layers.dropout(inputs=decoder, rate=dropout_rate, training=is_training, name='drop1')
      ## Add additional layers while accounting for input layer
      for layer in range(num_layers-2,-1,-1):
          decoder = tf.layers.dense(inputs=decoder,
                                    units=pow(2,hidden_unit_2power-layer),
                                    activation=tf.nn.relu,
                                    kernel_initializer=tf.glorot_uniform_initializer(),
                                    kernel_regularizer=tf.keras.regularizers.l2(l2_weight),
                                    use_bias=True,
                                    bias_initializer=init_ops.zeros_initializer(),
                                    name='fc'+str(num_layers-layer))
          if batch_norm: decoder = tf.layers.batch_normalization(decoder, training=is_training, renorm=batch_renorm, name='batch_norm_'+str(num_layers-layer))
          decoder = tf.layers.dropout(inputs=decoder, rate=dropout_rate, training=is_training, name='drop'+str(num_layers-layer))
      decoder_mean = tf.layers.dense(inputs=decoder,
                                units=output_size,
                                activation=None,
                                kernel_initializer=tf.glorot_uniform_initializer(),
                                kernel_regularizer=None,
                                use_bias=True,
                                bias_initializer=init_ops.zeros_initializer(),
                                name='fc'+str(num_layers+1))
      if FLAGS.decoder_variance == "per_sample":
          decoder_var = tf.layers.dense(inputs=decoder,
                                    units=1,
                                    activation=tf.nn.relu,
                                    kernel_initializer=tf.glorot_uniform_initializer(),
                                    kernel_regularizer=None,
                                    use_bias=True,
                                    bias_initializer=init_ops.zeros_initializer(),
                                    name='fc'+str(num_layers+1)+"_var")
          decoder = tfp.distributions.MultivariateNormalDiag(
                loc=decoder_mean,
                scale_diag=(tf.ones(output_size)*tf.nn.softplus(decoder_var)+1e-8),
                allow_nan_stats=False,
                validate_args=False,
                name="reconstructed_cell")
      else:
          decoder_var = tf.ones(1)
          decoder = tfp.distributions.MultivariateNormalDiag(
                loc=decoder_mean,
                scale_diag=tf.ones(output_size),
                allow_nan_stats=False,
                validate_args=False,
                name="reconstructed_cell")
  return decoder #, decoder_var



#! /usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys, os, glob, gzip, time
from functools import partial
from importlib import import_module

## Math
import numpy as np

## tensorflow imports
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.platform import app
from tensorflow.python.platform import flags

# def save_mixtures(self, sess, resultsObj, FLAGS, VAE, mixture_input_data, datatype, batch_size=100):
#     batch_size = np.amin((batch_size, mixture_input_data.shape[0]))
#     results=[]
#     ## Loop through bulk in batches
#     for i in range(0, mixture_input_data.shape[0], batch_size):
#         end_itr = np.amin((i + batch_size, mixture_input_data.shape[0]))
#         data_batch = mixture_input_data[i:end_itr,:]
#         ## Compute mixture from weighted components
#         for index, value in enumerate(self.current_celltypes, 0):
#             with tf.variable_scope(value, tf.AUTO_REUSE):
#                 emb      = sess.run(VAE[value].emb_saver_sample, feed_dict={VAE[value].fdata_ph: data_batch})
#                 purified = sess.run(VAE[value].rec_sampled_saver, feed_dict={VAE[value].emb_sampled_ph: emb[tf.newaxis,]})
#             ## Component specific contribution to mixture profile
#             test = sess.run(tf.gather(self.proportions[:,index,tf.newaxis], list(range(i,end_itr)), axis=0))
#             pure_data += sess.run(tf.multiply(purified, tf.gather(self.proportions[:,index,tf.newaxis], list(range(i,end_itr)), axis=0)))
#         results.append(pure_data)
#     resultsObj.mixture_saver(datatype, results)

def save_samples(sess,
                 resultsObj,
                 FLAGS,
                 VAE,
                 component_labels,
                 stage="prebatch"):
    if not os.path.exists(os.path.abspath(FLAGS.logdir + '/' + stage + '/sampled_results')):
        os.makedirs(os.path.abspath(FLAGS.logdir + '/' + stage + '/sampled_results'))
    ## Save component and mixture results
    for scope in np.unique(component_labels):
        print("Saving: "+scope)
        VAE[scope].save_sampled(sess=sess,
                                FLAGS=FLAGS,
                                resultsObj=resultsObj,
                                datatype=stage,
                                scope=scope,
                                dir=stage+"/sampled")

def save_results(sess,
                 resultsObj,
                 FLAGS,
                 VAE,
                 component_labels,
                 component_data,
                 hvg_masks,
                 marker_gene_masks,
                 stage="component",
                 component_valid_data=None,
                 component_valid_labels=None,
                 mixture_input_data=None):
    ## Log a bunch of results!
    if FLAGS.log_results is True:
        ## Create results directories
        if not os.path.exists(os.path.abspath(FLAGS.logdir + '/' + stage + '/component_train_results')):
            os.makedirs(os.path.abspath(FLAGS.logdir + '/' + stage + '/component_train_results'))
        if not os.path.exists(os.path.abspath(FLAGS.logdir + '/' + stage + '/component_valid_results')):
            os.makedirs(os.path.abspath(FLAGS.logdir + '/' + stage + '/component_valid_results'))
        if mixture_input_data is not None:
            if not os.path.exists(os.path.abspath(FLAGS.logdir + '/' + stage + '/mixture_results')):
                os.makedirs(os.path.abspath(FLAGS.logdir + '/' + stage + '/mixture_results'))
        ## Save component and mixture results
        for scope in np.unique(component_labels):
            with tf.variable_scope(scope, tf.AUTO_REUSE):
                print("Saving: "+scope)
                comp_data = component_data[np.where(component_labels == scope)[0],:]
                ## Component vae will only model hvgs specific to cell type
                if mixture_input_data is not None:
                    if hvg_masks is not None:
                        comp_data = comp_data[:,np.nonzero(hvg_masks[scope])[0]]
                        mixture_input_data_scope = mixture_input_data[:,np.nonzero(hvg_masks[scope])[0]]
                    else:
                        mixture_input_data_scope = mixture_input_data

                if comp_data.shape[0] > 0:
                    VAE[scope].save_results(sess=sess,
                                            resultsObj=resultsObj,
                                            datatype="component",
                                            mode="train",
                                            stage=stage,
                                            FLAGS=FLAGS,
                                            data=comp_data,
                                            scope=scope,
                                            dir=stage+"/component_train")
                if component_valid_data is not None and component_valid_data.shape[0] > 0:
                    VAE[scope].save_results(sess=sess,
                                            resultsObj=resultsObj,
                                            datatype="component",
                                            mode="test",
                                            stage=stage,
                                            FLAGS=FLAGS,
                                            data=component_valid_data[np.where(component_valid_labels == scope)[0],:],
                                            scope=scope,
                                            dir=stage+"/component_valid")
                if mixture_input_data is not None:
                    VAE[scope].save_results(sess=sess,
                                            resultsObj=resultsObj,
                                            datatype="purified",
                                            mode="train",
                                            stage=stage,
                                            FLAGS=FLAGS,
                                            data=mixture_input_data_scope,
                                            scope=scope,
                                            dir=stage+"/mixture")




#! /usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function ##

import sys, os, glob, gzip, time
from functools import partial
from importlib import import_module

## Math
import numpy as np

## tensorflow imports
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.platform import app
from tensorflow.python.platform import flags

##
from . import vaeBackend
from . import batchBackend
from . import deconvBackend
from . import utils
from . import saving_utils

def runVAE(resultsObj,
           FLAGS,
           config,
           component_data,
           component_labels,
           save_labels,
           mixture_data,
           marker_gene_masks=None,
           hvg_masks=None,
           max_steps_components=None,
           component_reconstruction_data=None,
           mixture_reconstruction_data=None,
           component_valid_data=None,
           component_valid_labels=None,
           mixture_valid_data=None,
           mixture_valid_labels=None,
           loadings=None):
    """Initializes and trains component VAE(s) and combined model.
    Args:
        FLAGS     (flags.FLAGS): Tensorflow flags.
        config    (ConfigProto): Hardware configuration options.
        component_data (matrix): Component VAE input data.
        component_labels  (str): Membership vector for each sample in component data.
        mixture_data   (matrix): Combined model input data to estimate proportions for.
        marker_gene_masks  (str): Genes to be used during deconvolution.
        component_reconstruction_data (matrix): Data to compare against component VAE reconstructions.
        mixture_reconstruction_data   (matrix): Data to compare against combined model reconstructions.
        component_valid_data (matrix): Component VAE validation (test) data.
        component_valid_labels  (str): Membership vector for each sample in validation (test) component data.
        mixture_valid_data (matrix): Component VAE validation (test) data.
        mixture_valid_labels  (str): Membership vector for each sample in validation (test) component data.
        loadings (matrix): Projection matrix to convert from low dimensions to gene expression space.
    Returns:
        proportions: Mixture proportions for each cell with respect to all component VAE(s).
        weights: Unnormalized mixture probabilities.
    """

    # ## Start deconvolution at top of tree (Python dictionary retains insertion order)
    # for key in component_labels.keys():
    ## Dictionary containing all VAE models named by unique elements within datasets
    VAE = {}
    ## Define network structure
    graph = tf.Graph()
    with graph.as_default():
        global_step_component = tf.train.get_or_create_global_step()
        ## Create component VAEs
        for scope in np.unique(component_labels):
            ## Set up encoder model. q(z|x)
            VAE[scope] = vaeBackend.modelVAE(FLAGS,
                                              latent_dims=FLAGS.latent_dims,
                                              input_size=component_data.shape[-1],
                                              output_size=component_data.shape[-1] if FLAGS.rec_project is False else component_reconstruction_data.shape[-1], ## Determines the output size of decoder
                                              num_samples=component_data[np.where(component_labels == scope)[0],:].shape[0],
                                              reconstruction=component_reconstruction_data,
                                              loadings=tf.cast(loadings, tf.float32) if loadings is not None else None,
                                              scope=scope,
                                              marker_masks=marker_gene_masks,
                                              hvg_masks=hvg_masks)

        ## Construct combined model for training/all mixture data which can potentially update VAEs
        if mixture_data is not None and FLAGS.deconvolution is True:

            deconvModel = deconvBackend.deconvModel(FLAGS=FLAGS,
                                                    VAE=VAE,
                                                    datasets=component_labels,
                                                    input_size=mixture_data.shape[-1],
                                                    output_size=mixture_data.shape[-1], ## Determines the output size of decoder
                                                    num_samples=mixture_data.shape[0],
                                                    reconstruction=mixture_data,
                                                    marker_gene_masks=marker_gene_masks,
                                                    loadings=tf.cast(loadings, tf.float32) if loadings is not None else None,
                                                    scope="deconvolution")

            batchModel = batchBackend.batchModel(FLAGS=FLAGS,
                                                 VAE=VAE,
                                                 mixture_weights=deconvModel.mixture_weights,
                                                 proportions=deconvModel.proportions,
                                                 datasets=component_labels,
                                                 input_size=mixture_data.shape[-1],
                                                 output_size=mixture_data.shape[-1], ## Determines the output size of decoder
                                                 num_samples=mixture_data.shape[0],
                                                 reconstruction=mixture_data,
                                                 loadings=tf.cast(loadings, tf.float32) if loadings is not None else None,
                                                 scope="batch_correction")

        ## Tensorboard summary and model checkpoint saving
        summary_op, summary_writer, summary_writer_test, saver = utils.summary_and_checkpoint(FLAGS, graph)

    ## Training scope
    with tf.Session(graph=graph, config=config) as sess:

        ## Set the logging level for tensorflow to only fatal issues
        tf.logging.set_verbosity(tf.logging.FATAL)

        ## Define seed at the graph-level
        ## From docs: If the graph-level seed is set, but the operation seed is not:
        ## The system deterministically picks an operation seed in conjunction with
        ## the graph-level seed so that it gets a unique random sequence.
        tf.set_random_seed(FLAGS.seed)

        ## Initialize everything
        tf.global_variables_initializer().run()
        print("Done random initialization")

        if not os.path.exists(os.path.abspath(FLAGS.logdir)):
           os.makedirs(os.path.abspath(FLAGS.logdir))

        ## Assert that nothing more can be added to the graph
        # tf.get_default_graph().finalize()

        ## Track epoch loss
        epoch_loss=[]

        ## Don't normalize output
        if component_reconstruction_data is None:
            component_reconstruction_data = component_data
        if mixture_reconstruction_data is None:
            mixture_reconstruction_data = mixture_data

        ## Initialize the component VAE Dataset iterators
        for scope in np.unique(component_labels):
            sess.run(VAE[scope].train_init_op, feed_dict={VAE[scope].data_ph: component_data[np.where(component_labels == scope)[0],:],
                                                          VAE[scope].rec_data_ph: component_reconstruction_data[np.where(component_labels == scope)[0],:]})

        ## Component training!
        if FLAGS.train_component is True:
            for step in range(1,FLAGS.max_steps_component+1):
                for scope in np.unique(component_labels):
                    if (step <= max_steps_components[scope]) and (VAE[scope].early_stop is not True):
                        with tf.variable_scope(scope, tf.AUTO_REUSE):
                            _, _, summaries_train, train_loss, logp, kl_loss, mse_loss = sess.run([VAE[scope].train_op, VAE[scope].update_op, VAE[scope].summary_op, VAE[scope].train_loss, VAE[scope].avg_distortion, VAE[scope].avg_KL_div, VAE[scope].mse],
                                                                                                  options=None,
                                                                                                  run_metadata=None)
                            ## Log mphate data
                            if FLAGS.m_phate_logging is True and step < 500:
                                embedding = sess.run(VAE[scope].emb_saver_sample, feed_dict={VAE[scope].fdata_ph: mixture_data})
                                resultsObj.update_mphate(scope, embedding)

                            if step % FLAGS.print_step == 0 or step == FLAGS.max_steps_component:
                                if FLAGS.component_corr_check is True:
                                    train_corr = sess.run(VAE[scope].corr)
                                    print("(%s) Step %s: %-6.2f    MSE: %-8.4f    LogP: %-8.4f    KLL %-8.4f    Corr %-8.4f" % (scope, step, train_loss, mse_loss, logp, kl_loss, train_corr))
                                else:
                                    print("(%s) Step %s: %-6.2f    MSE: %-8.4f    LogP: %-8.4f    KLL %-8.4f" % (scope, step, train_loss, mse_loss, logp, kl_loss))

                            if FLAGS.monitor_bulk_mse is True:
                                bulk_loss = VAE[scope].mixture_mse_monitor(mixture_data, FLAGS)
                                print("%-6.2f" % bulk_loss)

                            if component_valid_data is not None:
                                comp_valid_data = component_valid_data[np.where(component_valid_labels == scope)[0],:]
                                if comp_valid_data.shape[0] > 0 :
                                    ## Swap to testing data
                                    sess.run(VAE[scope].test_init_op, feed_dict={VAE[scope].data_ph: comp_valid_data,
                                                                                 VAE[scope].rec_data_ph: comp_valid_data})
                                    ## Test model
                                    if FLAGS.component_corr_check is True:
                                        summaries_test, test_loss, test_logp, test_kl_loss, test_mse_loss, test_corr = sess.run([VAE[scope].summary_op, VAE[scope].train_loss, VAE[scope].avg_distortion, VAE[scope].avg_KL_div, VAE[scope].mse, VAE[scope].corr])
                                    else:
                                        summaries_test, test_loss, test_logp, test_kl_loss, test_mse_loss, test_corr = sess.run([VAE[scope].summary_op, VAE[scope].train_loss, VAE[scope].avg_distortion, VAE[scope].avg_KL_div, VAE[scope].mse])
                                    ## Early stopping if enabled and minimum number of steps has been achieved
                                    if FLAGS.early_stopping is True and step >= FLAGS.early_min_step and FLAGS.component_corr_check is True:
                                        utils.early_stopping(FLAGS, VAE, scope, test_corr)

                                    ## Swap back to training data
                                    sess.run(VAE[scope].train_init_op, feed_dict={VAE[scope].data_ph: component_data[np.where(component_labels == scope)[0],:],
                                                                                  VAE[scope].rec_data_ph: component_reconstruction_data[np.where(component_labels == scope)[0],:]})
                            else:
                                    ## Early stopping if enabled and minimum number of steps has been achieved
                                    if FLAGS.early_stopping is True and step >= FLAGS.early_min_step and FLAGS.component_corr_check is True:
                                        train_corr = sess.run(VAE[scope].corr)
                                        utils.early_stopping(FLAGS, VAE, scope, train_corr)

                            ## Report loss
                            if step % FLAGS.log_step == 0 or step == FLAGS.max_steps_component:
                                summary_writer.add_summary(summaries_train, str(step))
                                summary_writer.flush()

                                ## Record results
                                resultsObj.update_component_metrics("train", scope, step, train_loss, mse_loss, logp, kl_loss)

                                ## Run test data if available
                                if component_valid_data is not None:
                                    if comp_valid_data.shape[0] > 0 :
                                        resultsObj.update_component_metrics("test", scope, step, test_loss, test_mse_loss, test_logp, test_kl_loss)
                                        ## Summary reports (Tensorboard), testing summary
                                        summary_writer_test.add_summary(summaries_test, str(step))
                                        summary_writer_test.flush()


                    ## Save model and summaries
                    if step % FLAGS.log_step == 0 or step == FLAGS.max_steps_component:
                        ## Summary reports (Tensorboard)
                        summary_writer.add_summary(summaries_train, str(step))
                        summary_writer.flush()
                        if component_valid_data is not None:
                            summary_writer_test.add_summary(summaries_test, str(step))
                            summary_writer_test.flush()
                        ## Write out graph
                        if step == FLAGS.max_steps_component:
                            save = saver.save(sess, os.path.abspath(FLAGS.logdir + '/ckpt/model_component.ckpt'), step)

            ## Save reconstruction results prior to mixture_stage training
            if FLAGS.log_results is True:
                saving_utils.save_results(sess=sess,
                                   resultsObj=resultsObj,
                                   FLAGS=FLAGS,
                                   VAE=VAE,
                                   component_labels=save_labels,
                                   component_data=component_data,
                                   hvg_masks=hvg_masks,
                                   marker_gene_masks=marker_gene_masks,
                                   stage="component",
                                   component_valid_labels=component_valid_labels,
                                   component_valid_data=component_valid_data,
                                   mixture_input_data=mixture_data)
            # if FLAGS.log_samples is True:
            #     saving_utils.save_samples(sess=sess,
            #                        resultsObj=resultsObj,
            #                        FLAGS=FLAGS,
            #                        VAE=VAE,
            #                        component_labels=component_labels,
            #                        stage="prebatch")
        else:
            ## Load a previously saved model
            saver.restore(sess, tf.train.latest_checkpoint(os.path.abspath(FLAGS.logdir +  '/ckpt')))

        ## Proportion training
        if mixture_data is not None and FLAGS.deconvolution is True:

            ## Initialize the mixture dataset
            sess.run(deconvModel.mix_iter_data.initializer, feed_dict={deconvModel.mix_data_ph: mixture_data})

            ## Initialize the mixture dataset
            sess.run(batchModel.mix_iter_data.initializer, feed_dict={batchModel.mix_data_ph: mixture_data})

            for step in range(1,FLAGS.proportion_steps+1):
                ## Proportion optimizer
                _, summaries, prop_loss = sess.run([deconvModel.train_proportion_op, summary_op, deconvModel.proportion_loss])

                if step % FLAGS.print_step == 0 or step == FLAGS.proportion_steps:
                    print("(Proportions) Step: %s    MSE: %-8.4f" % (step, prop_loss))

                ## Report loss
                if step % FLAGS.log_step == 0:
                    summary_writer.add_summary(summaries, str(step+FLAGS.max_steps_component))

                if step % np.ceil(mixture_data.shape[0]/FLAGS.batch_size_mixture) == 0:
                    epoch_loss = np.mean(epoch_loss)
                    resultsObj.update_mixture_metrics("train", epoch_loss, epoch_loss)
                    epoch_loss = [] ## Reset
                else:
                    epoch_loss.append(prop_loss)

                ## Save model and summaries
                if step % FLAGS.log_step == 0 or step == FLAGS.proportion_steps:
                    ## Summary reports (Tensorboard)
                    summary_writer.add_summary(summaries, str(step+FLAGS.max_steps_component))
                    ## Get estiamted proportions
                    proportions = sess.run(deconvModel.proportions)
                    #weights = sess.run(deconvModel.mixture_weights)
                    resultsObj.update_proportions(step, proportions)
                    if step == FLAGS.proportion_steps:
                        ## Write out graph
                        save = saver.save(sess, os.path.abspath(FLAGS.logdir + '/ckpt/model_combined.ckpt'), step+FLAGS.max_steps_component)

            ## Save mixture probabilities
            # proportions = sess.run(deconvModel.proportions)
            # np.savetxt(FLAGS.logdir + "/probabilities.csv",
            #            proportions,
            #            delimiter=",",
            #            header=','.join(str(x) for x in np.unique(component_labels)),
            #            comments='')

            ## Save mixture weights
            # weights = sess.run(deconvModel.mixture_weights)
            # np.savetxt(FLAGS.logdir + "/weights.csv",
            #            weights,
            #            delimiter=",",
            #            header=','.join(str(x) for x in np.unique(component_labels)),
            #            comments='')

            # deconvModel.save_mixtures(sess=sess,
            #                             resultsObj=resultsObj,
            #                             FLAGS=FLAGS,
            #                             VAE=VAE,
            #                             mixture_input_data=mixture_data,
            #                             datatype='prebatch',
            #                             batch_size=100)

            ## Mixture batch-correction training
            if FLAGS.batch_correction is True:
                ## Record the VAE weights after pretraining
                # if FLAGS.tied_model_l2 is True and step == FLAGS.max_steps_component:
                #     for scope in np.unique(component_labels):
                #         print("Assigning weights")
                #         VAE[scope].assign_network_weights(sess=sess)

                ## Train!
                for step in range(1,FLAGS.max_steps_combined+1):
                    ## Combined optimizer
                    _, summaries, train_loss, mse_loss, marker_mse, non_marker_mse = sess.run([batchModel.train_batch_op, summary_op, batchModel.train_batch_loss, batchModel.batch_loss, batchModel.marker_mse, batchModel.non_marker_mse])

                    if step % FLAGS.print_step == 0 or step == FLAGS.max_steps_combined:
                        print("(Batch Correct) Step %s: %-6.2f    LogP(mixture): %-8.4f    MSE(marker): %-8.4f    MSE(Non-marker): %-8.4f" % (step, train_loss, mse_loss, marker_mse, non_marker_mse))

                    ## Report loss
                    if step % FLAGS.log_step == 0:
                        for scope in np.unique(component_labels):
                            with tf.variable_scope(scope, tf.AUTO_REUSE):
                                train_loss, logp, kl_loss, mse_loss = sess.run([VAE[scope].train_loss, VAE[scope].avg_distortion, VAE[scope].avg_KL_div, VAE[scope].mse], options=None, run_metadata=None)
                                resultsObj.update_component_metrics("train-post", scope, step, train_loss, mse_loss, logp, kl_loss)
                                #print(sess.run(VAE[scope].data_index))
                        summary_writer.add_summary(summaries, str(step+FLAGS.max_steps_component+FLAGS.proportion_steps))
                        resultsObj.update_mixture_metrics("train", train_loss, mse_loss)

                    ## Save model and summaries
                    if step % FLAGS.log_step == 0 or step == FLAGS.max_steps_combined:
                        ## Summary reports (Tensorboard)
                        summary_writer.add_summary(summaries, str(step+FLAGS.max_steps_component+FLAGS.proportion_steps))
                        ## Get estiamted proportions
                        proportions = sess.run(batchModel.proportions)
                        #weights = sess.run(batchModel.mixture_weights)
                        resultsObj.update_proportions(step, proportions)
                        if step == FLAGS.max_steps_combined:
                            ## Write out graph
                            save = saver.save(sess, os.path.abspath(FLAGS.logdir + '/ckpt/model_combined.ckpt'), step+FLAGS.max_steps_component+FLAGS.proportion_steps)

                if FLAGS.log_results is True:
                    print("Logging combined results")
                    ## Save reconstruction results prior to mixture_stage training
                    saving_utils.save_results(sess=sess,
                                       resultsObj=resultsObj,
                                       FLAGS=FLAGS,
                                       VAE=VAE,
                                       component_labels=save_labels,
                                       component_data=component_data,
                                       hvg_masks=hvg_masks,
                                       marker_gene_masks=marker_gene_masks,
                                       stage="combined",
                                       component_valid_labels=component_valid_labels,
                                       component_valid_data=component_valid_data,
                                       mixture_input_data=mixture_data)

                # deconvModel.save_mixtures(sess=sess,
                #                             resultsObj=resultsObj,
                #                             FLAGS=FLAGS,
                #                             VAE=VAE,
                #                             mixture_input_data=mixture_data,
                #                             datatype='postbatch',
                #                             batch_size=100)

            # print(deconvModel.trainable_prop)
            # print(deconvModel.trainable_comb)
        #print(tf.get_collection(tf.GraphKeys.UPDATE_OPS))
        ## Record the VAE weights after pretraining
        # if FLAGS.tied_model_l2 is True:
        #     for scope in np.unique(component_labels):
        #         print("Assigning weights")
        #         VAE[scope].assign_network_weights(sess=sess, assign=False)
    ## Return depending on deconvolution
    return resultsObj



#! /usr/bin/env python
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

## Basics
import sys, os, glob, gzip, time, traceback

## Verbosity of tensorflow output. Filters: (1 INFO) (2 WARNINGS) (3 ERRORS)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

## Math
import numpy as np

## tensorflow imports
# import tensorflow as tf
import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import logging
import numpy as np
from tensorflow.python.platform import app
from tensorflow.python.platform import flags
from sklearn.model_selection import StratifiedShuffleSplit

from . import utils
from . import vaeTrain

class deconvModel(object):
    def __init__(self, component_data, component_label, mixture_data, save_label=None):
        ## Required parameters
        self.component_data  = component_data
        self.component_label = component_label
        self.mixture_data    = mixture_data
        if save_label is None:
            self.save_label = self.component_label
        else:
            self.save_label = save_label

        ## Deconvolution
        self.celltypes = np.unique(self.component_label)

        ## Log model state and save locations
        self.results_dir = None
        self.flags       = None

        ## train, test split
        self.train_index = None
        self.test_index  = None

        ## Initialize results class
        self.deconvResults = utils.deconvResult(celltypeNames = self.celltypes)

    def deconvolve( self,
                    ## Masking
                    marker_gene_mask = None,
                    hvg_mask         = None,
                    ## Learning rates (Careful changing these)
                    component_learning_rate = 1e-3,
                    proportion_learning_rate = 1e-3,
                    combined_learning_rate  = 1e-4,
                    decay_lr = False,
                    ## Network architecture
                    num_dr = 50,
                    num_latent_dims = 32,
                    num_layers = 3,
                    hidden_unit_2power = 9,
                    batch_norm_layers = False,
                    decoder_var = 'per_sample',
                    latent_x = False,
                    seed = 1234,
                    ## Early stopping
                    early_stopping = True,
                    early_min_step = 100,
                    max_patience = 100,
                    ## VAE training parameters
                    max_steps_component  = 5000,
                    batch_size_component = 100,
                    corr_check           = False,
                    ## Combined training parameters
                    deconvolution       = True,
                    batch_correction    = True,
                    max_steps_combined = 10000,
                    batch_size_mixture   = 100,
                    max_steps_proportion = 1000,
                    ## KL term
                    KL_weight            = 1.0,
                    marker_weight        = 1,
                    ## MCMC
                    num_monte_carlo      = 15,
                    ## L2 Reg term
                    tied_model_l2        = True,
                    ## Training Setup
                    training_method = 'train',
                    num_folds       = 1,
                    heldout_percent = 0.33,
                    ## Hardware params
                    cuda_device = 0,
                    ## Logging params
                    log_step        = 5000,
                    print_step      = 100,
                    log_results     = True,
                    save_results    = False,
                    log_samples     = True,
                    m_phate_logging = False,
                    ## Path params
                    logdir     = './tmp/',
                    save_to_object = True,
                    save_to_disk   = False,
                    model_name = 'default' ):

        ## Try to run deconv, if fail, be sure to clean up tensorflow properly.
        try:
            ## Setup annotation and masks
            component_labels,  \
            marker_gene_masks, \
            hvg_masks,         \
            max_steps_components = utils.setup_annotations_masks(self.component_label,
                                                                 marker_gene_mask,
                                                                 hvg_mask,
                                                                 max_steps_component,
                                                                 self.celltypes)
            ## Tensorflows variant of argparse
            FLAGS = flags.FLAGS

            ## Result flags
            flags.DEFINE_string('logdir', os.path.abspath(logdir), 'Save directory.')
            flags.DEFINE_string('model', model_name, 'Name of data to use for saving.')

            ## Saving flags
            flags.DEFINE_boolean('log_samples', str(log_samples), 'To log samples from each VAE')
            flags.DEFINE_boolean('m_phate_logging', str(m_phate_logging), 'Logging for m-phate viz')
            flags.DEFINE_boolean('log_results', str(log_results), 'To log detailed results and model files')
            flags.DEFINE_boolean('save_results', str(save_results), 'To save detailed results and model files')
            flags.DEFINE_boolean('save_to_object', str(save_to_object), 'Purified data is saved to results object')
            flags.DEFINE_boolean('save_to_disk', str(save_to_disk), 'Purified data is saved out to disk')

            ## Reporting
            flags.DEFINE_integer('log_step', log_step, 'When to output model summaries.')
            flags.DEFINE_integer('print_step', print_step, 'When to print out sumamries.')
            flags.DEFINE_boolean('monitor_bulk_mse', False, "Should bulk mse be reported during scVAE training")

            ## Prior
            flags.DEFINE_integer('num_dr', num_dr, 'Number of dimensions prior to network.')

            ## Component model architecture flags
            flags.DEFINE_integer('latent_dims', str(num_latent_dims), 'dim of latent Z')
            flags.DEFINE_float('KL_weight', KL_weight, 'Weight on KL term of elbo.')
            flags.DEFINE_boolean('KL_warmup', True, 'Slowly increase KL_weight during training.')
            flags.DEFINE_integer('num_monte_carlo', num_monte_carlo, 'number of samples to estimate likelihood')
            flags.DEFINE_string('decoder_variance', decoder_var, 'Structure of decoder distribution') ## "per_gene", "per_sample", 'default (unit variance)'
            flags.DEFINE_boolean('latent_x', latent_x, 'Should the purified data be treated as latent or not')

            ## Early stopping flags
            flags.DEFINE_boolean('early_stopping', early_stopping, 'Should early stopping be performed.')
            flags.DEFINE_integer('early_min_step', early_min_step, 'How many steps before early stopping procedure starts.')
            flags.DEFINE_integer('max_patience', max_patience, 'How long to wait before early stopping.')

            flags.DEFINE_float('component_learning_rate', component_learning_rate, 'Initial learning rate for VAE training.')
            flags.DEFINE_float('proportion_learning_rate', proportion_learning_rate, "Initial learning rate for proportion est.")
            flags.DEFINE_float('combined_learning_rate', combined_learning_rate, 'Initial learning rate for VAE correction.')
            flags.DEFINE_float('min_learning_rate', 1e-8, 'Minimum learning rate.')
            flags.DEFINE_boolean('decay_lr', str(decay_lr), 'Should learning rate be decayed')
            flags.DEFINE_float('decay_rate', 0.3, 'How fast to decay learning rate.')
            flags.DEFINE_integer('decay_step', 1000, 'Decay interval')

            flags.DEFINE_integer('max_steps_component', str(max(max_steps_components.values())), 'maximum number of traning iterations')
            flags.DEFINE_integer('max_steps_combined', str(max_steps_combined), 'maximum number of traning iterations')
            flags.DEFINE_integer('batch_size_component', str(batch_size_component), 'size of minibatch')
            flags.DEFINE_integer('batch_size_mixture', str(batch_size_mixture), 'size of minibatch')
            flags.DEFINE_boolean('batch_norm', str(batch_norm_layers), 'To include batch_norm layers in model')

            flags.DEFINE_boolean('kl_analytic', True, 'Built in or manual KL computation, current version of tensorflow probability has a bug. Set to False.')
            flags.DEFINE_integer('seed', str(seed), 'random seed for reproducability')

            ## Data specific flags
            flags.DEFINE_boolean('deconvolution', deconvolution, 'Should deconvolution be performed')
            flags.DEFINE_boolean('rec_project', False, 'Will pc space be used as input')
            flags.DEFINE_boolean('L2_norm_data', False, 'Will perform gene-wise l2 normalization.')

            ## Maskings
            flags.DEFINE_string('component', None, 'Subset to just one component (for debugging)')
            flags.DEFINE_string('component_remove', None, 'Remove specific component from reference data.')

            flags.DEFINE_boolean('tpm_softmax', False, 'Should TPM data be rescaled instead of reconstruction to actual values')
            flags.DEFINE_integer('tpm_scale_factor', 10000, 'Scale factor for TPM')

            ## Component model training
            flags.DEFINE_boolean('train_component', True, 'Should component network weights be (re)trained.')
            flags.DEFINE_integer('num_layers', str(num_layers), "Number of encoder/decoder NN layer.")
            flags.DEFINE_integer('hidden_unit_2power', str(hidden_unit_2power), "Starting number of hidden units in the first layer.")
            flags.DEFINE_boolean('tied_model_l2', tied_model_l2, 'L2 regularization between prior and current model weights')

            ## Combined model flags
            flags.DEFINE_integer('proportion_steps', str(max_steps_proportion), 'How long to train combined model before updating VAE(s)')
            flags.DEFINE_boolean('mixture_softmax', True, 'Should softmax be used to compute mixture probabilities.')
            flags.DEFINE_boolean('combined_corr_check', corr_check, 'Should pearson correlation of mixture and reconstructed mixture be recorded? (Slow)')
            flags.DEFINE_boolean('component_corr_check', corr_check, 'Should pearson correlation of mixture and reconstructed mixture be recorded? (Slow)')

            ## Combined loss weights
            flags.DEFINE_boolean('batch_correction', batch_correction, 'Should combined network update components')
            flags.DEFINE_integer('marker_weight', marker_weight, 'Initial learning rate for VAE training.')

            ## regularizer options (UNUSED)
            flags.DEFINE_float('l1_reg_weight', 1e-5, 'Weight of l1 regularizer for combined model.')
            flags.DEFINE_string('input_weight_reg', 'None', 'Should the input gene specific weights have a regularizer')   ## Sparse adjustment
            flags.DEFINE_string('output_weight_reg', 'None', 'Should the output geme specific weights have a regularizer') ## Sparse adjustment
            flags.DEFINE_boolean('mix_weight_reg', False, 'Should the mixture weights have a regularizer') ## Mixture weights

            ## validation setup
            flags.DEFINE_string('training_method', training_method, 'How many, if any, folds to perform CV')
            flags.DEFINE_integer('kfold_validation', str(num_folds), "How many, if any, folds to perform CV")
            flags.DEFINE_float('heldout_percent', str(heldout_percent), "How much data to holdout for testing")

            ## HARDWARE
            flags.DEFINE_string('cuda_device', str(cuda_device), 'Select the GPU for this job')

            ## Remove some logging
            logging.getLogger('tensorflow').setLevel(logging.FATAL)

            ## Hardware configurations
            config = utils.define_hardware(FLAGS)

            ## Setup saving directory
            FLAGS = utils.initialize_model_dir(FLAGS)

            ## Write out all run options for reproducability
            utils.write_flags(FLAGS)

            ## Adjust data based on training mode
            if FLAGS.training_method == "train":
                deconvResults = vaeTrain.runVAE(resultsObj = self.deconvResults,
                                                 FLAGS=FLAGS,
                                                 config=config,
                                                 component_data=self.component_data,
                                                 component_labels=component_labels,
                                                 save_labels=self.save_label,
                                                 mixture_data=self.mixture_data,
                                                 marker_gene_masks=marker_gene_masks,
                                                 hvg_masks=hvg_masks,
                                                 max_steps_components=max_steps_components,
                                                 component_reconstruction_data=None,
                                                 mixture_reconstruction_data=self.mixture_data,
                                                 component_valid_data=None,
                                                 component_valid_labels=None,
                                                 mixture_valid_data=None,
                                                 mixture_valid_labels=None,
                                                 loadings=None)
            elif FLAGS.training_method == "validate":
                ## Setup kfold iterator object
                split_iter = StratifiedShuffleSplit(n_splits=2, test_size=FLAGS.heldout_percent, random_state=FLAGS.seed)
                for train_index, test_index in split_iter.split(self.component_data, component_labels):
                    deconvResults = vaeTrain.runVAE(resultsObj = self.deconvResults,
                                                     FLAGS=FLAGS,
                                                     config=config,
                                                     component_data=self.component_data[train_index,:],
                                                     component_labels=component_labels[train_index],
                                                     save_labels=self.save_label,
                                                     mixture_data=self.mixture_data,
                                                     marker_gene_masks=marker_gene_masks,
                                                     hvg_masks=hvg_masks,
                                                     max_steps_components=max_steps_components,
                                                     component_reconstruction_data=None,
                                                     mixture_reconstruction_data=self.mixture_data,
                                                     component_valid_data=self.component_data[test_index,:],
                                                     component_valid_labels=component_labels[test_index],
                                                     mixture_valid_data=None,
                                                     mixture_valid_labels=None,
                                                     loadings=None)
                    self.train_index = train_index
                    self.test_index  = test_index
            else:
                print("Mode not recognized, leaving data as is.")

            ## Final logging
            self.results_dir = FLAGS.logdir
            self.flags = tf.app.flags.FLAGS.flag_values_dict()
        except:
            print("An error occured: ")
            traceback.print_exc()
        finally:
            ## Clean up
            FLAGS.remove_flag_values(FLAGS.flag_values_dict())

    def check_args(self):
        print("test")



"""
Copyright 2018 Quonlab.
Copyright 2016 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import sys, os

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import clip_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.training import optimizer as tf_optimizer
from tensorflow.python.training import training_util

from . import architectures

class modelVAE(object):
  def __init__(self, FLAGS, latent_dims, input_size, output_size, num_samples, reconstruction, loadings, scope, marker_masks=None, hvg_masks=None):
    ## Info
    print("Constructing VAE: "+scope)
    print(" |")

    ## Create component VAE model
    with tf.variable_scope(scope, tf.AUTO_REUSE):
        ## Params for VAE
        self.scope       = scope
        self.latent_dims = latent_dims
        self.input_size  = input_size
        self.output_size = np.count_nonzero(hvg_masks[self.scope]) if hvg_masks is not None else input_size
        self.num_samples = num_samples
        self.batch_size  = np.amin((FLAGS.batch_size_component, self.num_samples))

        ## Maskings
        self.marker_mask = marker_masks[self.scope] if marker_masks is not None else None
        self.non_marker_mask = np.ones(self.output_size)-self.marker_mask if marker_masks is not None else None
        self.hvg_mask    = hvg_masks[self.scope] if hvg_masks is not None else None

        ## Params for early stopping
        self.early_stop  = False
        self.patience    = 0
        self.best_val_loss = 1e20

        ## L2_tied
        self.tied_model_l2 = tf.constant(value=0, dtype=tf.float32)

        ## Architectures (parameterizing dists)
        self.encoder_func = tf.make_template('encoder',
                                             architectures.vae_encoder,
                                             FLAGS=FLAGS,
                                             ndim=self.latent_dims,
                                             num_layers=FLAGS.num_layers,
                                             hidden_unit_2power=FLAGS.hidden_unit_2power,
                                             scope=self.scope,
                                             batch_norm=FLAGS.batch_norm,
                                             create_scope_now_=False)
        self.decoder_func = tf.make_template('decoder',
                                             architectures.vae_decoder,
                                             FLAGS=FLAGS,
                                             output_size=self.output_size,
                                             num_layers=FLAGS.num_layers,
                                             hidden_unit_2power=FLAGS.hidden_unit_2power,
                                             scope=self.scope,
                                             batch_norm=FLAGS.batch_norm,
                                             create_scope_now_=False)

        self.latent_prior = architectures.vae_prior(ndim=self.latent_dims, scope=self.scope)
        self.reconstruction_prior = architectures.vae_prior(ndim=self.output_size, scope=self.scope)

        ## Basics
        self.global_step = tf.train.get_or_create_global_step()
        self.step = tf.Variable(1, name='step', trainable=False, dtype=tf.int32)

        ## Set up inputs. Mini-batch
        with tf.name_scope("component_data"):
            self.index         = tf.range(start=0, limit=self.num_samples, dtype=tf.int32)
            self.data_ph       = tf.placeholder(tf.float32, (None, self.input_size))
            self.rec_data_ph   = tf.placeholder(tf.float32, (None, self.input_size))
            self.train_dataset = tf.data.Dataset.from_tensor_slices((self.data_ph, self.index)).shuffle(self.num_samples).repeat().batch(self.batch_size).prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
            self.test_dataset  = tf.data.Dataset.from_tensor_slices((self.data_ph, self.index)).batch(tf.shape(self.data_ph, out_type=tf.dtypes.int64)[0]).prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
            self.iterator      = tf.data.Iterator.from_structure(self.train_dataset.output_types,
                                                                 self.train_dataset.output_shapes)
            # self.iterator     = self.dataset.make_initializable_iterator()
            self.data_batch, self.data_index = self.iterator.get_next()
            ## Choice of datasets
            self.train_init_op = self.iterator.make_initializer(self.train_dataset)
            self.test_init_op = self.iterator.make_initializer(self.test_dataset)

        ## Holds purified bulk data, helper for combined model
        with tf.name_scope("purified_data"):
            self.pure_data = None

        ## Compute embeddings and logits.
        self.emb = self.encoder_func(inputs=self.data_batch, is_training=True)
        self.latent_sample = self.emb.sample(FLAGS.num_monte_carlo)
        self.rec = self.decoder_func(inputs=self.latent_sample, is_training=True)
        if FLAGS.tpm_softmax is True:
            self.rec_data = tf.nn.softmax(tf.reduce_mean(self.rec.mean(), axis=0), axis=-1) * FLAGS.tpm_scale_factor
        else:
            #self.rec_data = tf.reduce_mean(self.rec.mean(), axis=0)
            self.rec_data = tf.reduce_mean(self.rec.mean(), axis=0)

        ## Set up KL weight warmup
        with tf.name_scope("KL_warmup"):
            if FLAGS.KL_warmup is True: self.KL_weight = tf.cast(tf.minimum(
                                            FLAGS.KL_weight*(tf.cast(self.step, dtype=tf.float32)/(FLAGS.max_steps_component/tf.constant(2.0, dtype=tf.float32))),
                                            FLAGS.KL_weight),
                                            dtype=tf.float32)
            else: self.KL_weight = FLAGS.KL_weight
            tf.compat.v1.summary.scalar("KL_weight", self.KL_weight)
        tf.compat.v1.summary.scalar("training_step", self.step)
        ## Monitor patience counter
        tf.compat.v1.summary.scalar("Patience", self.patience)

        ## Add VAE specific loss
        self.add_ELBO_loss(FLAGS)

        ## Metrics
        self.mse  = tf.losses.mean_squared_error(self.rec_data, self.data_batch, loss_collection="metrics")
        tf.compat.v1.summary.scalar(self.scope+"_component_mse", self.mse)

        ## Correlation for tracking training performance
        if FLAGS.component_corr_check is True:
            self.corr = tfp.stats.correlation(self.rec_data, self.data_batch, sample_axis=-1, event_axis=0)
            self.corr = tf.reduce_mean(tf.linalg.tensor_diag_part(self.corr), axis=0)
            tf.compat.v1.summary.histogram(scope+"_component_correlation", self.corr)

        ## Set up learning rate
        if FLAGS.decay_lr is True:
            with tf.name_scope("learning_rate"):
                self.t_learning_rate = tf.maximum(
                    tf.train.exponential_decay(
                        FLAGS.component_learning_rate,
                        self.step,
                        FLAGS.decay_step,
                        FLAGS.decay_rate),
                    FLAGS.min_learning_rate)
        else:
            self.t_learning_rate = FLAGS.component_learning_rate

        ## Create training operation
        self.update_op=tf.compat.v1.get_collection(tf.GraphKeys.UPDATE_OPS, scope=self.scope)
        self.create_train_op(self.t_learning_rate, update_ops=self.update_op)

        ## Batch correction loss terms
        if self.marker_mask is not None:
            self.compute_batch_loss(FLAGS)

        ## Record VAE encoder dist. parameters
        # tf.compat.v1.summary.histogram(scope+"_mean", self.emb.mean())
        # tf.compat.v1.summary.histogram(scope+"_variance", self.emb.variance())

        # ## Record VAE decoder dist. parameters
        # tf.compat.v1.summary.histogram(scope+"_mean_decoder", self.rec.mean())
        # tf.compat.v1.summary.histogram(scope+"_variance_decoder", self.rec.variance())

        ## Define weight saving
        #self.record_network_weights()
        #self.tied_weight_regularizer()

        ## Define saving operations
        self.define_saver_ops(FLAGS=FLAGS, is_training=False)
        self.define_inverse_saver_ops(FLAGS=FLAGS, is_training=True)
        self.define_summary_ops()

  def add_logp_loss(self, FLAGS):
    with tf.variable_scope('LogLikelihood'):
        ## `distortion` is the negative log likelihood.
        self.distortion = -self.rec.log_prob(self.data_batch)
        self.avg_distortion = tf.reduce_mean(input_tensor=self.distortion)
        tf.compat.v1.summary.scalar("Reconstruction", self.avg_distortion)

  def add_KL_loss_encoder(self, FLAGS):
    with tf.variable_scope('KL'):
        self.rate = tfp.distributions.kl_divergence(self.emb, self.latent_prior, allow_nan_stats=False)
        self.avg_KL_div = tf.reduce_mean(input_tensor=self.rate)
        tf.compat.v1.summary.scalar("KL_encoder", self.avg_KL_div)

  def add_KL_loss_decoder(self, FLAGS):
    with tf.variable_scope('KL_decoder'):
        self.rate_decoder = tfp.distributions.kl_divergence(self.rec, self.reconstruction_prior, allow_nan_stats=False)
        self.avg_KL_div_decoder = tf.reduce_mean(input_tensor=self.rate_decoder)
        tf.compat.v1.summary.scalar("KL_decoder", self.avg_KL_div_decoder)

  def add_ELBO_loss(self, FLAGS, smoothing=0.0):
    with tf.variable_scope('ELBO'):
        ## Components of elbo
        self.add_logp_loss(FLAGS)
        self.add_KL_loss_encoder(FLAGS)
        ## Per sample loss
        if FLAGS.latent_x is True:
            self.add_KL_loss_decoder(FLAGS)
            elbo_local = -((self.KL_weight * self.rate) + (self.KL_weight * self.rate_decoder) + self.distortion)
        else:
            elbo_local = -((self.KL_weight * self.rate) + self.distortion)
        ## Mean loss for batch
        self.elbo = tf.reduce_mean(elbo_local)
        tf.losses.add_loss(-self.elbo,
                           loss_collection=tf.GraphKeys.LOSSES)
        tf.compat.v1.summary.scalar("elbo", -self.elbo)

  def create_train_op(self, learning_rate, update_ops=None, check_numerics=True, summarize_gradients=False):
    """Create and return training operation."""
    with tf.control_dependencies(update_ops):
        with tf.name_scope('Optimizer'):
            self.l2_reg = tf.losses.get_regularization_loss(scope=self.scope)
            #self.train_loss = tf.add_n(tf.losses.get_losses(loss_collection=tf.GraphKeys.LOSSES, scope=self.scope) + tf.losses.get_regularization_losses(scope=self.scope))
            self.train_loss = tf.add_n(tf.losses.get_losses(loss_collection=tf.GraphKeys.LOSSES, scope=self.scope) + tf.losses.get_regularization_losses(scope=self.scope))
            ## Minimize loss function
            optimizer = tf.train.AdamOptimizer(learning_rate)
            self.train_op = optimizer.minimize(loss=self.train_loss, global_step=self.step, var_list=tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope))
            ## Monitor
            #tf.compat.v1.summary.scalar('Learning_Rate_component', learning_rate)
            #tf.compat.v1.summary.scalar('Loss_Total_component', self.train_loss)
            #tf.compat.v1.summary.scalar('Loss_Regularization', self.l2_reg)
            return(self.train_op)

  ## Operation to compute batch correction loss
  def compute_batch_loss(self, FLAGS):
      with tf.variable_scope(self.scope, tf.AUTO_REUSE):
        ## Reconstruction of markers: marker_mask
        marker_measured = tf.multiply(self.data_batch, self.marker_mask)
        marker_reconstruction = tf.multiply(tf.reduce_mean(self.rec.mean(), axis=0), self.marker_mask)
        ## Reweight marker priority using partial observations
        # if FLAGS.marker_weight > 1:
        #     marker_measured = tf.tile(marker_measured, (FLAGS.marker_weight, 1))
        #     marker_reconstruction = tf.tile(marker_reconstruction, (FLAGS.marker_weight, 1))
        self.marker_mse = tf.losses.mean_squared_error(marker_measured,
                                                       marker_reconstruction,
                                                       weights=1.0,
                                                       scope=self.scope)
        ## Reconstruction of non_markers: np.ones - marker_mask
        if np.sum(self.non_marker_mask) > 0:
            non_marker_measured = tf.boolean_mask(self.data_batch, self.non_marker_mask)
            non_marker_reconstruction = tf.multiply(tf.reduce_mean(self.rec.mean(), axis=0), self.non_marker_mask)
            self.non_marker_mse = tf.losses.mean_squared_error(non_marker_measured,
                                                               non_marker_reconstruction,
                                                               weights=1.,
                                                               scope=self.scope)
        else:
            self.non_marker_mse = 0.0

  # ## Operation to compute batch correction loss
  # def compute_batch_loss(self, FLAGS):
  #     with tf.variable_scope(self.scope, tf.AUTO_REUSE):
  #       ## Reconstruction of markers: marker_mask
  #       marker_measured = tf.boolean_mask(self.data_batch, self.marker_mask, axis=1)
  #       marker_reconstruction = tf.boolean_mask(tf.reduce_mean(self.rec.sample(FLAGS.num_monte_carlo)[0], axis=0), self.marker_mask, axis=1)
  #       ## Reweight marker priority using partial observations
  #       if FLAGS.marker_weight > 1:
  #           marker_measured = tf.tile(marker_measured, (FLAGS.marker_weight, 1))
  #           marker_reconstruction = tf.tile(marker_reconstruction, (FLAGS.marker_weight, 1))
  #       self.marker_mse = tf.losses.mean_squared_error(marker_measured,
  #                                                      marker_reconstruction,
  #                                                      weights=1.0,
  #                                                      scope=self.scope)
  #       ## Reconstruction of non_markers: np.ones - marker_mask
  #       if np.sum(self.non_marker_mask) > 0:
  #           non_marker_measured = tf.boolean_mask(self.data_batch, self.non_marker_mask, axis=1)
  #           non_marker_reconstruction = tf.boolean_mask(tf.reduce_mean(self.rec.sample(FLAGS.num_monte_carlo)[0], axis=0), self.non_marker_mask, axis=1)
  #           self.non_marker_mse = tf.losses.mean_squared_error(non_marker_measured,
  #                                                              non_marker_reconstruction,
  #                                                              weights=1.,
  #                                                              scope=self.scope)
  #       else:
  #           self.non_marker_mse = 0.0

  # def record_network_weights(self):
  #     ## Get all weights from current VAE and save just their names
  #     self.weight_names  = [tensor.name for tensor in tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=self.scope) if tensor.name.endswith('kernel:0')]
  #     self.saved_weights = {}
  #     ## Holds all the ops required to "fill", fixed weight matrices after training VAEs.
  #     for name in self.weight_names:
  #       weight_matrix = tf.get_default_graph().get_tensor_by_name(name)
  #       self.saved_weights[name] = tf.Variable(tf.zeros(weight_matrix.shape), shape=weight_matrix.shape)
  #       #tf.compat.v1.summary.histogram("fixed_"+name[:-2], self.saved_weights[name])

  def assign_network_weights(self, sess, assign=True):
      ## Hardcopy the weight matrices from dense layers
      for name in self.weight_names:
          print("RECORDING: "+name)
          if assign is True:
              self.saved_weights[name].load(sess.run(name))
              np.savetxt('/home/ucdnjj/matrix_files/'+name[:-2].replace("/","_")+'.csv', sess.run(self.saved_weights[name]), delimiter=',')
          else:
              np.savetxt('/home/ucdnjj/matrix_files_end/'+name[:-2].replace("/","_")+'.csv', sess.run(self.saved_weights[name]), delimiter=',')

  # def tied_weight_regularizer(self):
  #     with tf.name_scope('tied_model_l2'):
  #         for name in self.weight_names:
  #             self.tied_model_l2 += tf.nn.l2_loss(tf.subtract(tf.get_default_graph().get_tensor_by_name(name), self.saved_weights[name]))
  #         tf.compat.v1.summary.scalar("tied_model_l2", self.tied_model_l2)

  def define_summary_ops(self):
      ## VAE's summary set
      self.summary_op = tf.summary.merge_all(scope=self.scope)

  ## Defines copies of the network that take in placeholders for final pass of compelete dataset
  def define_saver_ops(self, FLAGS, is_training):
    ## Saver helper function for encoder
    self.fdata_ph = tf.placeholder(tf.float32, (None, self.input_size))
    self.emb_saver = self.encoder_func(inputs=self.fdata_ph, is_training=is_training)
    self.emb_saver_mean = self.emb_saver.mean()
    self.emb_saver_var  = self.emb_saver.variance()
    self.emb_saver_sample = self.emb_saver.sample(FLAGS.num_monte_carlo)[0]

    ## Saver helper function for decoder
    self.emb_ph = tf.placeholder(tf.float32, (None, self.latent_dims))
    self.rec_saver = self.decoder_func(inputs=self.emb_ph, is_training=is_training)
    self.rec_saver_lp = self.rec_saver.log_prob(self.fdata_ph)

    ## Saver helper function for sampled decoder
    self.emb_sampled_ph = tf.placeholder(tf.float32, (None, None, self.latent_dims))
    self.rec = self.decoder_func(inputs=self.emb_sampled_ph, is_training=is_training)
    #self.rec_var  = tf.reduce_mean(self.rec_var, axis=0)
    self.rec_mean = tf.reduce_mean(self.rec.mean(), axis=0)

  def define_inverse_saver_ops(self, FLAGS, is_training=True):
      self.inv_emb_saver = self.encoder_func(inputs=self.fdata_ph, is_training=is_training)
      self.inv_emb_saver_sample = self.inv_emb_saver.sample(FLAGS.num_monte_carlo)[0]
      self.inv_rec = self.decoder_func(inputs=self.emb_sampled_ph, is_training=is_training)
      self.inv_rec_mean = tf.reduce_mean(self.inv_rec.mean(), axis=0)

  def mixture_mse_monitor(self, data, FLAGS, is_training=False):
      emb_comp_temp = sess.run(self.emb_saver_sample, feed_dict={self.fdata_ph: data})
      reconstruction = sess.run(self.rec_mean, feed_dict={self.emb_ph: emb_comp_temp})
      return sess.run(tf.compat.v1.losses.mean_squared_error(data,  reconstruction))

  ## Evaluate endpoint op in batchs of data.
  def save_sampled(self, sess, resultsObj, datatype, scope, FLAGS, dir, nsample=100):
      with tf.variable_scope(self.scope, tf.AUTO_REUSE):
        samples = sess.run(self.latent_prior.sample(nsample))
        rec_comp = sess.run(self.rec_sampled_saver, feed_dict={self.emb_sampled_ph: samples[tf.newaxis,]})
        resultsObj.sample_saver(datatype, scope, samples, rec_comp)
        if FLAGS.save_results is True:
            np.savetxt(FLAGS.logdir + "/"+dir+"_results/reconstructions.csv", rec_comp, delimiter=",")

  ## Save scRNA outputs
  def save_results(self, sess, resultsObj, datatype, mode, stage, FLAGS, data, scope, dir, batch_size=100):
    ## Handle large mixture input
    rec_comp = []; emb_comp  = []; rec_comp_lprob = [];  var_comp  = [];
    with tf.variable_scope(self.scope, tf.AUTO_REUSE):
        for i in range(0, data.shape[0], batch_size):
            data_batch = data[i:i + batch_size]
            emb_comp_temp = sess.run(self.emb_saver_sample, feed_dict={self.fdata_ph: data_batch})
            emb_comp.append(emb_comp_temp)
            rec_comp_lprob.append(sess.run(self.rec_saver_lp, feed_dict={self.emb_ph: emb_comp_temp, self.fdata_ph: data_batch}))
            rec_comp.append(sess.run(self.rec_mean, feed_dict={self.emb_sampled_ph: emb_comp_temp[tf.newaxis,]}))
            #var_comp.append(sess.run(self.rec_var, feed_dict={self.emb_sampled_ph: emb_comp_temp[tf.newaxis,]}))

    ## Combined batched results
    emb_comp = np.concatenate(emb_comp)
    rec_comp_lprob = np.concatenate(rec_comp_lprob)
    rec_comp = np.concatenate(rec_comp)
    #if FLAGS.decoder_variance != 'fixed':
    #    var_comp = np.concatenate(var_comp)
    ##
    # inv_rec_comp = np.concatenate(inv_rec_comp)

    if FLAGS.save_to_object is True:
        resultsObj.update_saver(datatype, mode, stage, scope, rec_comp, emb_comp, rec_comp_lprob, var_comp)

    if FLAGS.save_to_disk is True:
        print("Not setup")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import sys, os

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import tensorflow_probability as tfp
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import clip_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.training import optimizer as tf_optimizer
from tensorflow.python.training import training_util

class batchModel(object):
    def __init__(self, FLAGS, VAE, datasets, input_size, output_size, num_samples, proportions, mixture_weights, reconstruction=None, loadings=None, VAE_parameters=None, scope="batch_correction"):
        ## Params for deconvolution model
        self.input_size  = input_size
        self.output_size = output_size
        self.num_samples = num_samples
        self.batch_size  = np.amin((FLAGS.batch_size_mixture, self.num_samples)) ## Batch size should not be larger than dataset size
        ## Organizing name for cmobined model
        self.scope = scope
        ## Cell types in model
        self.current_celltypes = np.unique(datasets)
        self.num_component = len(self.current_celltypes)
        ## Normalize mixture weights
        self.mixture_weights = mixture_weights
        self.proportions = proportions

        ## Combined model steps
        self.step = tf.Variable(1, name='batch_step', trainable=False, dtype=tf.int32)

        ## Build combined model
        with tf.variable_scope(self.scope, tf.AUTO_REUSE):
            ## Create parameters and metadata for batch correction
            self.build_dataset();
            self.build_model();
            self.batch_correction(FLAGS, VAE);
            ## MSE between reconstruction and measured mixture during proportion est. or batch correction
            self.add_loss();
            ## Record VAE trainable parameters
            self.trainable_params += tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "bulk_distribution")

            # ## Corr
            # if FLAGS.combined_corr_check is True:
            #     self.corr = tfp.stats.correlation(self.mix_data_batch, self.mixture_rec, sample_axis=-1, event_axis=0)
            #     self.corr = tf.reduce_mean(tf.linalg.tensor_diag_part(self.corr), axis=0)
            #     tf.compat.v1.summary.scalar("Mixture_correlation", self.corr)

            ## Build the optimizer
            self.create_train_op(FLAGS)

    def build_dataset(self):
        with tf.name_scope("mixture_data"):
            ## Mixture indexing
            self.index = tf.range(start=0, limit=self.num_samples, dtype=tf.int32)
            ## Dataset
            self.mix_data_ph = tf.placeholder(tf.float32, (None, self.input_size))
            self.mix_dataset = tf.data.Dataset.from_tensor_slices((self.mix_data_ph, self.index)).shuffle(self.num_samples).repeat().batch(self.batch_size)
            self.mix_iter_data  = self.mix_dataset.make_initializable_iterator()
            self.mix_data_batch, self.data_index  = self.mix_iter_data.get_next()

    def build_model(self):
        ## Hold results of linear reconstitution of mixture
        self.mixture_rec = tf.zeros([self.batch_size, self.output_size], tf.float32)
        ## Batch correction loss
        self.marker_mse     = tf.constant(0., tf.float32)
        self.non_marker_mse = tf.constant(0., tf.float32)
        self.KL_vae         = tf.constant(0., tf.float32)
        self.elbo_vae       = tf.constant(0., tf.float32)
        self.regularization_vae = tf.constant(0., tf.float32)
        self.regularization_tied = tf.constant(0., tf.float32)

    def batch_correction(self, FLAGS, VAE):
        ## Compute mixture from weighted components
        for index, value in enumerate(self.current_celltypes, 0):
            with tf.variable_scope(value, tf.AUTO_REUSE):
                self.data_emb = VAE[value].encoder_func(self.mix_data_batch, is_training=False).sample(FLAGS.num_monte_carlo)
                self.data_dist = VAE[value].decoder_func(self.data_emb, is_training=False)
                self.data_rec = tf.reduce_mean(self.data_dist.mean(), axis=0)

            with tf.name_scope("LinearCombo"):
                ## Component specific contribution to mixture profile
                VAE[value].pure_data = tf.multiply(self.data_rec, tf.gather(self.proportions[:,index,tf.newaxis], self.data_index, axis=0))
                ## Reconstruct mixture celltype-wise
                self.mixture_rec = tf.add(self.mixture_rec, VAE[value].pure_data)
                ## Record metrics
                # if VAE[value].marker_mask is not None:
                #     self.marker_mse      += VAE[value].marker_mse
                #     self.non_marker_mse  += VAE[value].non_marker_mse
                self.KL_vae              += (VAE[value].avg_KL_div * VAE[value].KL_weight)
                self.elbo_vae            += VAE[value].elbo
                self.regularization_vae  += VAE[value].l2_reg
                #self.regularization_tied += VAE[value].tied_model_l2
                tf.compat.v1.summary.scalar('KL_'+ VAE[value].scope, self.KL_vae)
                tf.compat.v1.summary.scalar('ELBO_'+VAE[value].scope, -self.elbo_vae)
                tf.compat.v1.summary.scalar('L2_REG_'+ VAE[value].scope, self.regularization_vae)

            ## Gather VAE variances for bulk variance calculation
            try:
                #self.variances_vae += [tf.ones((10,100))]
                self.trainable_params += tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, VAE[value].scope)
            except AttributeError:
                #self.variances_vae = [tf.ones((10,100))]
                self.trainable_params = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, VAE[value].scope)
            #tf.summary.histogram(value + "_weights", self.mixture_weights[:,index])
            # tf.summary.histogram(value + "_probs", self.proportions[:,index])

        ## Define bulk distribution for proportion estimation
        with tf.name_scope("bulk_distribution"):
            # if FLAGS.decoder_variance == "per_sample":
            #     merged_vae_vars = tf.concat(self.variances_vae, axis=0)
            #     merged_vae_vars = tf.tile(merged_vae_vars[tf.newaxis,], [self.batch_size,1])
            # else:
            #     merged_vae_vars = tf.stack(self.variances_vae, axis=-1)
            #     merged_vae_vars = tf.tile(merged_vae_vars, [self.batch_size,1])
            # print(merged_vae_vars)
            # print(merged_vae_vars)
            gathered_probs  = tf.gather(self.proportions, self.data_index, axis=0)
            #print(gathered_probs)
            # combined_var_input = tf.concat([gathered_probs, merged_vae_vars], axis=1)
            # print(combined_var_input)
            # combined_var_input.set_shape((self.batch_size, self.num_component+(self.num_component*self.batch_size)))
            # print(combined_var_input)
            self.dist_var = tf.layers.dense(inputs=gathered_probs, #combined_var_input, #tf.tile(tf.reshape(merged_vae_vars, [-1])[tf.newaxis,], [self.batch_size,1])], axis=1),  #[tf.newaxis,]
                                      units=1,
                                      activation=None,
                                      kernel_initializer=tf.glorot_uniform_initializer(),
                                      kernel_regularizer=tf.nn.relu,
                                      use_bias=True,
                                      bias_initializer=init_ops.zeros_initializer(),
                                      name='fc_var')
            self.mixture_rec_dist = tfp.distributions.MultivariateNormalDiag(
                                    loc=self.mixture_rec,
                                    scale_diag=(tf.ones(self.output_size)*tf.nn.softplus(self.dist_var)+1e-8),
                                    allow_nan_stats=False,
                                    validate_args=True,
                                    name="reconstructed_cell_dist")
            ## summaries
            # tf.compat.v1.summary.histogram("prop_bulk_dist_mean", self.mixture_rec_dist.mean())
            # tf.compat.v1.summary.histogram("prop_bulk_dist_var", self.mixture_rec_dist.variance())

    ## Bulk mse loss
    def add_loss(self):
        ## Log probability
        self.batch_loss = -tf.reduce_mean(self.mixture_rec_dist.log_prob(self.mix_data_batch))
        tf.summary.scalar("batch_loss", self.batch_loss)

    ## Create operating to adjust VAEs with fixed proportions
    def create_train_op(self, FLAGS):
        """Create and return training operation."""
        with tf.name_scope('Optimizer_batch'):
            ## Set up learning rate
            if FLAGS.decay_lr is True:
                with tf.name_scope("learning_rate_proportion"):
                    self.batch_learning_rate = tf.maximum(
                        tf.train.exponential_decay(
                            FLAGS.combined_learning_rate,
                            self.step,
                            FLAGS.decay_step,
                            FLAGS.decay_rate),
                        FLAGS.min_learning_rate)
            else:
                self.batch_learning_rate = FLAGS.combined_learning_rate

            ## Minimize loss function
            optimizer = tf.train.AdamOptimizer(self.batch_learning_rate)

            ## Just optimizing proportions based on marker expr. recontruction of bulk
            self.train_batch_loss = self.batch_loss + (-self.elbo_vae) + (self.KL_vae) + tf.reduce_sum(self.regularization_vae)
            if FLAGS.tied_model_l2 is True:
                self.train_batch_loss += tf.reduce_sum(self.regularization_tied)

            ## If markers are provided then we can include marker vs. non-marker mse terms at different magnitudes
            # if FLAGS.marker_weight > 1:
            #     self.train_batch_loss += self.marker_mse + self.non_marker_mse

            ## Batch correction optimization
            self.train_batch_op = optimizer.minimize(loss=self.train_batch_loss, global_step=self.step, var_list=self.trainable_params)

            ## Monitor
            tf.summary.scalar('Loss_Total', self.train_batch_loss)



import setuptools

setuptools.setup(name='scProjection',
    version='0.22',
    description='Projection and Deconvolution using deep heirarchical and generative neural network.',
    url='https://github.com/ucdavis/quonlab/tree/master/development/deconvAllen',
    author='Nelson Johansen, Gerald Quon',
    author_email='njjohansen@ucdavis.edu, gquon@ucdavis.edu',
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'tensorflow',
        'tensorflow_probability',
        'sklearn',
        'numpy'
      ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6")
